/*
 * The functions in this file
 * implement commands that search in the
 * forward and backward directions. There are
 * no special characters in the search strings.
 * Probably should have a regular expression
 * search, or something like that.
 */
#include	<stdio.h>
#include	"ed.h"

#if AtST
/* overlay "me2"	dal: can't do this in Alcyon C */
#endif

/* mb: added: */
#define	eq(c,d,s)	((d)=='?' || (s ? (FOLD(c)==(d)) : ((c)==(d))))
#define	FOLD(c)		tolower(c)

/* mb: completely rewritten.
 * Search forward.
 * Get a search string from the
 * user, and search, beginning at ".",
 * for the string. If found, reset the
 * "." to be just after the match string,
 * and [perhaps] repaint the display.
 * Bound to "M-S".
 */
forwsearch(f, n)
	{
	register LINE *clp;
	register char *pp;
	register int  cbo;
	register char c;
	register int  s;
	register int  found;
	LINE *tlp;
	int  tbo;

	if ((s=readpattern("Forward search"))==ABORT || pat[0]=='\0')
		return(ABORT);
	clp = curwp->w_dotp;
	cbo = curwp->w_doto;
	found = FALSE;
	pp = pat;
	for(;;) 
		{
		if (cbo >= llength(clp)) 
			{
			clp = lforw(clp);
			if (clp == curbp->b_linep) 
				{
				found = FALSE;
				break;
				}
			cbo = 0;
			c = '\n';
			}
		else 
			{
			c = lgetc(clp, cbo++);
			}
		if (*pp == 0x17) 				/* ctrl-W */
			{
			if (c==' ' || c=='\t' || c=='\n')
				continue;
			pp++;
			if (*pp == '\0') 
				{
				if (cbo)
					--cbo;
				break;
				}
			}
		if (! eq(c, *pp, s)) 
			{
			if (found) 
				{
				found = FALSE;
				pp = pat;
				clp = tlp;
				cbo = tbo;
				}
			continue;
			}
		if (! found) 
			{
			found = TRUE;
			tlp = clp;
			tbo = cbo;
			}
		if (*(++pp) == '\0')
			break;
		}   /* end of for(;;) */
	if (found) 
		{
		curwp->w_dotp = clp;
		curwp->w_doto = cbo;
		curwp->w_flag |= WFMOVE;
		mlerase();
		}
	else 
		{
		mlwrite("Not found");
		}
	return (found);
	}

/* mb: completely rewritten.
 * Reverse search.
 * Get a search string from the
 * user, and search, starting at "."
 * and proceeding toward the front of
 * the buffer. If found "." is left
 * pointing at the first character of
 * the pattern [the last character that
 * was matched]. Bound to "M-R".
 */
backsearch(f, n)
	{
	register LINE *clp;
	register char *pp;
	register int  cbo;
	register char c;
	register int  s;
	register int  found;
	LINE *tlp;
	char *epp;
	int  tbo;

	if ((s=readpattern("Reverse search"))==ABORT || pat[0]=='\0')
		return(ABORT);
	clp = curwp->w_dotp;
	cbo = curwp->w_doto;
	epp = pat;
	while (*(epp+1) != '\0')
		++epp;
	found = FALSE;
	pp = epp;
	for(;;) 
		{
		if (--cbo < 0) 
			{
			clp = lback(clp);
			if (clp == curbp->b_linep) 
				{
				found = FALSE;
				break;
				}
			cbo = llength(clp);
			c = '\n';
			}
		else 
			{
			c = lgetc(clp, cbo);
			}
		if (*pp == 0x17) 				/* ctrl-W */
			{
			if (pp == pat)
				break;
			if (c==' ' || c=='\t' || c=='\n')
				continue;
			--pp;
			}
		if (! eq(c, *pp, s)) 
			{
			if (found) 
				{
				found = FALSE;
				pp = epp;
				clp = tlp;
				cbo = tbo;
				}
			continue;
			}
		if (! found) 
			{
			found = TRUE;
			tlp = clp;
			tbo = cbo;
			}
		if (--pp < pat)
			break;
		}   /* end of for(;;) */
	if (found) 
		{
		curwp->w_dotp = clp;
		curwp->w_doto = cbo;
		curwp->w_flag |= WFMOVE;
		mlerase();
		}
	else 
		{
		mlwrite("Not found");
		}
	return (found);
	}

/* mb: eq() unnecessary now, deleted */

static	int	no_case = FALSE;
/*
 * Read a pattern.
 * Stash it in the external
 * variable "pat". The "pat" is
 * not updated if the user types in
 * an empty line. If the user typed
 * an empty line, and there is no
 * old pattern, it is an error.
 * Display the old pattern, in the
 * style of Jeff Lomicka. There is
 * some do-it-yourself control
 * expansion.
 */
readpattern(prompt)
	char	*prompt;
	{
	register char	*cp1;
	register char	*cp2;
	register int	c;
	register int	i;	/* mb: added */
	char	tpat[NPAT+21];
	int	s;

	cp1 = &tpat[0];
	if (pat[0] != '\0') 			/* Old pattern		*/
		{
		cp2 = &pat[0];
		while ((c = *cp2++) != 0) 
			{
			if (cp1 < &tpat[NPAT+20-6]) 	/* "??]: \0"	*/
				{
				if (c == '\n') 
					{
					*cp1++ = '<';
					c = '<';
					}
				else if (c == '\t') 
					{
					*cp1++ = '>';
					c = '>';
					}
				else if (c == 0x17) 
					{
					*cp1++ = '_';
					c = '_';
					}
				else if (iscntrl(c)) /* (c<0x20 || c==0x7F) */
					{
					*cp1++ = '^';
					c ^= 0x40;
					}
				else if (c == '%')	/* Map "%" to	*/
					*cp1++ = c;	/* "%%".	*/
				*cp1++ = c;
				}
			}
		*cp1++ = '\0';
		cp2 = tpat;
		}
	else
		cp2 = NULL;
	s = mlgets(prompt, cp2, tpat, NPAT, "\003\033");  /* Read pattern */
	if (s == ABORT)
		return (s);
	s = (s == '\003');
	i = 0;
	if (tpat[0] != '\0') 		/* specified */
		{
		no_case = s;			/* save case sensitivity */
		do 
			{
			c=tpat[i];
			pat[i] = (s ? FOLD(c) : c);
			i++;
			}
			while (c != '\0');
		pat[i] = tpat[i];		/* the terminating char */
		}
	c = pat[strlen(pat)+1];
	return (no_case);
	}

/*
 * mb: Search forward for complementing brace.
 * If not on an opening brace, search forwards
 * for one, then for corresponding closing brace.
 * Braces are {}, () and []. Braces inside quotes
 * (single or double) and comments are ignored.
 * Bound to "M-]", "M-)" and "M-}".
 */
forw_brace(f, n)
	{
	register char	ch;
	register char	oldch;
	register int	doto;
	register LINE	*dotp;
	register int	quoted;
	register int	comment;
	int	depth, status;
	char	older, quote;
	char	stack[128];

	status = TRUE;
	quoted = FALSE;
	comment = FALSE;
	depth = 0;
	dotp = curwp->w_dotp;
	doto = curwp->w_doto;
	ch = '\n';
	while (depth < 127) 
		{
		if ( doto >= llength(dotp) ) 
			{
			dotp = lforw(dotp);
			if (dotp == curbp->b_linep) 
				{
				status = ABORT;
				break;
				}
			doto = 0;
			continue;
			}
		older = oldch;
		oldch = ch;	
		ch = lgetc(dotp, doto++);
		if (quoted) 
			{
			if (ch == quote)
				if ( oldch!='\\'
				||  (oldch=='\\' && older=='\\'))
					quoted = FALSE;
			continue;
			}
		if (comment) 
			{
			if (ch == '/' && oldch == '*')
				comment = FALSE;
			continue;
			}
		if ((ch=='"' || ch=='\'') && oldch != '\\') 
			{
			quoted = TRUE;
			quote = ch;
			continue;
			}
		if (ch == '*' && oldch == '/') 
			{
			comment = TRUE;
			continue;
			}
		if (ch == '{')
			stack[depth++] = '}';
		else if (ch == '(')
			stack[depth++] = ')';
		else if (ch == '[')
			stack[depth++] = ']';
		else if ((ch=='}' || ch==')' || ch==']') && depth>0 ) 
			{
			if (stack[--depth] != ch) 
				{
				status = FALSE;
				break;
				}
			if (depth == 0)
				break;    /* found it! */
			}
		} 				/* end of while() */
	if (status == TRUE) 
		{
		curwp->w_dotp  = dotp;
		curwp->w_doto  = doto - 1;
		curwp->w_flag |= WFMOVE;
		return (TRUE);
		}
	if (status == FALSE || depth > 0)
		mlwrite("Braces not balanced!");
	return (status);
	}

/*
 * mb: Search backwards for complementing brace.
 * If not on a closing brace, search backwards
 * for one, then for corresponding opening brace.
 * Braces are {}, () and []. Braces inside quotes
 * (single or double) and comments are ignored.
 * Bound to "M-[", "M-(" and "M-{".
 */
back_brace(f, n)
	{
	register char	ch;
	register char	oldch;
	register int	doto;
	register LINE	*dotp;
	register int	quoted;
	register int	comment;
	int	depth, status;
	char	older, quote;
	char	stack[128];

	status = TRUE;
	quoted = FALSE;
	comment = FALSE;
	depth = 0;
	dotp = curwp->w_dotp;
	doto = curwp->w_doto + 1;	/* read THIS char, but: */
	if (doto > llength(dotp))	/* if at far eol...	*/
		--doto;			/* don't read garbage	*/
	ch = '\n';
	while (depth < 127) 
		{
		if ( doto <= 0 ) 
			{
			dotp = lback(dotp);
			if (dotp == curbp->b_linep) 
				{
				status = ABORT;
				break;
				}
			doto = llength(dotp);
			continue;
			}
		older = oldch;
		oldch = ch;	
		ch = lgetc(dotp, --doto);
		if (quoted) 
			{
			if (ch != '\\' && oldch == quote) 
				{
				oldch = '\n';
				quoted = FALSE;
				}
			else
				continue;
			}
		if (comment) 
			{
			if (ch == '/' && oldch == '*')
				comment = FALSE;
			continue;
			}
		if ((oldch=='"' || oldch=='\'') && ch!='\\') 
			{
			quoted = TRUE;
			quote = oldch;
			continue;
			}
		if ((older=='"' || older=='\'')
		    && ch=='\\' && oldch=='\\') 
 			{
			quoted = TRUE;
			quote = older;
			continue;
 			}
		if (ch == '*' && oldch == '/') 
			{
			comment = TRUE;
			continue;
			}
		if (ch == '}')
			stack[depth++] = '{';
		else if (ch == ')')
			stack[depth++] = '(';
		else if (ch == ']')
			stack[depth++] = '[';
		else if ((ch=='{' || ch=='(' || ch=='[') && depth>0 ) 
			{
			if (stack[--depth] != ch) 
				{
				status = FALSE;
				break;
				}
			if (depth == 0)
				break;    /* found it! */
			}
		} 				/* end of while() */
	if (status == TRUE) 
		{
		curwp->w_dotp  = dotp;
		curwp->w_doto  = doto;
		curwp->w_flag |= WFMOVE;
		return (TRUE);
		}
	if (status == FALSE || depth > 0)
		mlwrite("Braces not balanced!");
	return (status);
	}
