{A program to read an ASCII file and display, on the monitor,
a text which has fewer words per line.  
This permits easier reading by a humanoid.
                              version 1.0
                           by Merlin L. Hanson}
PROGRAM NarrowText;
{$I GEMSUBS.PAS}
{$I NARRWTXT.I}  {Contains CONST declaration}
  {Characters without ASCII graphics.}
  HT = 09;
  LF = 10;
  CR = 13;
  UpArrow   = $4800;
  DownArrow = $5000;
  Help      = $6200;
TYPE
  T1    = PACKED ARRAY[0..100000]OF char;  {An ASCII, or text file}
  T116  = FILE OF T1;
  TMode = (Narrow,Wide);
  T19   = (LineFeed,CarriageReturn,Tab,Mundane);
  TKindSrh = (FirstSearch,RepeatedSearch);
VAR
  Printer : text;
    
{$I AUXSUBS.PAS}

FUNCTION ReadNoEcho:char;
  VAR ThisCh:long_integer;

  FUNCTION RdNoEcho:long_integer;
    GEMDOS($07);

  BEGIN {ReadNoEcho}
    ThisCh := RdNoEcho;
    {Move scan code to lower byte.}
    IF ThisCh & $FF = 0
      THEN ReadNoEcho := CHR(INT(SHR(ThisCh,8)))
      ELSE ReadNoEcho := CHR(INT(ThisCh));
  END {readnoecho};    

PROCEDURE DoIt;
  VAR
    InFile : T116;                            {The input text file}
    NumberFileBytes : long_integer;           {File size from OS}
    Mode    : TMode;                          {Narrow or wide}
    InIndex : long_integer;                   {Offest in the file}
    Used    : integer; 
    Stack   : ARRAY[0..4000]OF long_integer;  {Handle files of up 100,000
                                              bytes with as few as 25 characters
                                              per line.}
    StackIndex : integer;                     {Next entry goes here}
    LastLine   : integer;                     {Number of 'new lines' in the file.}
    NbrChPerLine : integer;                   {New line width}
    MyBox        : dialog_ptr;                {for communal use}
  
  FUNCTION GetLineWidth:integer;
    {This is provided solely to provide an easy place for the user to 
    find, and edit, the desired line width.}
    VAR Width:long_integer;
    BEGIN
      Width := $AAAAAA32;   {50 characters per line}
      GetLineWidth := Width - $AAAAAA00;
    END {getlinewidth};

  PROCEDURE GetFile(VAR FileVar:T116);
    {Let the user select an input file.  Get it assigned as FileVar.}
    VAR MyPath,MyFile:string;

    {$I GETPATH.PAS}
    {$I GETFLSZ.PAS}

    BEGIN
      GetDriveAndPath(MyPath);
      MyPath := CONCAT(MyPath,'*.*');
      MyFile := '';
      GoToXY(2,35);
      InverseVideo;  Write(' Input File ');  NormVideo;
      IF Get_In_File(MyPath,MyFile)    
        THEN 
          BEGIN
            ClrScrn;
            RESET(FileVar,MyFile);
          END  
        ELSE HALT; {user cancel}    
      ClrScrn;
      {The first (and only) component of the file has been read.}
      GetFileSize(MyFile,NumberFileBytes);
    END {getfile};  

  PROCEDURE AppendEOF_Text;
    {Mimic the output produced by TOS on 'Show' file.}
    VAR S:string; i:integer;
    BEGIN
      S := CONCAT(#13,#10,#13,#10,'---END OF FILE---',#13,#10);
      FOR i := 0 TO 24 DO
        BEGIN
          InFile^[NumberFileBytes] := S[i];
          NumberFileBytes := NumberFileBytes+1;
        END;
    END {appendeof_text};
      
  PROCEDURE Patience;
    BEGIN
      Find_Dialog(MyWaitBx,MyBox);
      Center_Dialog(MyBox);
      Show_Dialog(MyBox,MyWaitBx);
    END {patience};

  PROCEDURE ShowHelp;
    VAR Sp:string; junk:integer;
    BEGIN {showhelp}
      ClrScrn;
      Find_Dialog(HelpBox,MyBox);
      Center_dialog(MyBox);
      junk := Do_Dialog(MyBox,0);
      Obj_SetState(MyBox,junk,Normal,TRUE);
      ClrScrn;
    END {showhelp};
    
  PROCEDURE ShowAbout;  
    VAR junk:integer;
    BEGIN
      ClrScrn;
      Find_Dialog(CredtBox,MyBox);
      Center_Dialog(MyBox);
      junk := Do_Dialog(MyBox,0);
      Obj_SetState(MyBox,junk,Normal,TRUE);
    END {showabout};

  PROCEDURE ProcessLine(FileIndex:long_integer; 
                       VAR ChUsed:integer;
                            VAR S:string);
    {Form a narrow line of text out of a wide (potentially infinite)
    line.  Replace HT, CR AND LF characters with suitable replacements,
    typically spaces or nulls.  When a line of the target width has
    been formed, return.  The target is to make a line containing
    NbrCharPerLine.}
    VAR 
      GoodCh,six,SpaceIndex : integer;  
      P_Graph : boolean;
      NewLine : boolean;
    
    FUNCTION FindLastSpace(LastIx:integer; VAR SpacePtr:integer):boolean;
      {Find the last space character in S. Set Space Ptr to
      point at it and return TRUE.  If there is no space, return FALSE}
      BEGIN
        SpacePtr := LastIx;
        WHILE (SpacePtr > 1) AND (S[SpacePtr] <> ' ') DO
          SpacePtr := SpacePtr-1;
        IF SpacePtr > 1
          THEN FindLastSpace := TRUE
          ELSE FindLastSpace := FALSE;  
      END {findlastspace};

    FUNCTION Paragraph(i:long_integer):boolean;  
      {See if there is any text prior to the preceding CR. 
      If there is return FALSE, otherwise return TRUE.
      i is the index in the file of
      a CR that has just been encountered.}
      VAR CR_Next,MyText :boolean;
      BEGIN
        CR_Next := FALSE;
        MyText := FALSE;
        i := i-1;
        IF i > 0
          THEN
            REPEAT
              IF (InFile^[i] > ' ') AND (InFile^[i] <= '~')
                THEN MyText := TRUE
                ELSE
                  IF (InFile^[i] = CHR(13))
                    THEN CR_Next := TRUE;
              i := i-1;      
            UNTIL (FileIndex = 0) OR MyText OR CR_Next;
        IF CR_Next
          THEN Paragraph := TRUE
          ELSE Paragraph := FALSE;
      END {paragraph};
      
    PROCEDURE EditString;
      {A string, S, contains a full (narrow) line of text.
      See if it can be broken into parts without distortion.}
      BEGIN 
        IF FindLastSpace(GoodCh,SpaceIndex)
          THEN
            BEGIN
              S[0] := CHR(SpaceIndex);
              ChUsed := ChUsed - (GoodCh-SpaceIndex);
            END
          ELSE {There was no space.  Print the entire line.}
            S[0] := CHR(GoodCh);    
      END {editstring};
      
    FUNCTION CharKind(ch:char):T19;
      BEGIN
        CASE ORD(ch) OF
          LF : CharKind := LineFeed;
          CR : CharKind := CarriageReturn;
          HT : CharKind := Tab;
          OTHERWISE : CharKind := Mundane;
        END {case};  
      END {charkind};
      
    FUNCTION HaveLine:boolean;      
      {If part of a line has been formed return TRUE.  
      Otherwise return FALSE.}
      BEGIN
        IF GoodCh > 0 
          THEN HaveLine := TRUE
          ELSE HaveLine := FALSE;
      END {haveline};

    PROCEDURE DoParagraph;
      {The sequence CR LF CR LF is taken to mean a new paragraph.
      Create a blank line in this case.}
      BEGIN
        P_Graph := TRUE;
        IF HaveLine 
          THEN
            BEGIN {\'uncount' this CR}
              FileIndex := FileIndex-1;
              ChUsed := ChUsed-1;
            END
          ELSE S := ' ';
      END {doparagraph};

    PROCEDURE DoCR;  
      BEGIN
        IF (S[six-1] = ' ') OR (six = 1)
          THEN {nil} 
          ELSE
            BEGIN
              S[six] := ' ';
              six := six+1;
              GoodCh := GoodCh+1;
            END;  
      END {docr};

    PROCEDURE ProcessTextRapidly;  
      {Go through text as rapidly as possible.  Most of the time delay
      after the file read is consumed in this procedure.}
      VAR ChProcessed:integer; fix:long_integer;
      BEGIN
        NewLine := FALSE;
        fix := FileIndex;
        REPEAT
          (*Write('six:', six, 'fix:', fix);*)
          S[six] := InFile^[fix];
          six := six+1;
          fix := fix+1;
        UNTIL (six > NbrChPerLine) OR (InFile^[fix] < ' ');
        ChProcessed := INT(fix-FileIndex);
        GoodCh := GoodCh + ChProcessed; 
        FileIndex := fix-1;                 {will be incremented later}
        ChUsed := ChUsed + ChProcessed-1;   {will be incremented later}
      END {processtextrapidly};
      
    BEGIN {processline}
      NewLine := FALSE;
      ChUsed := 0;
      GoodCh := 0;
      six    := 1;  {StringIndex}
      P_Graph := FALSE;
      S := '';
      REPEAT
        CASE CharKind(InFile^[FileIndex]) OF
          LineFeed : NewLine := TRUE;
          Mundane : 
            BEGIN
              IF NewLine AND (InFile^[FileIndex] = ' ') {discard leading blanks}
                THEN {nil}
                ELSE
                  BEGIN
                    IF (FileIndex+100 < NumberFileBytes)
                      THEN ProcessTextRapidly
                      ELSE
                        BEGIN
                          NewLine := FALSE;
                          S[six] := InFile^[FileIndex];
                          six    := six+1;
                          GoodCh := GoodCh+1;
                        END;  
                  END;  
            END;  
          Tab : {Replace HT with 1 space.}
            BEGIN
              S[six] := ' ';
              six := six+1;
              GoodCh := GoodCh+1;
            END;  
          CarriageReturn : 
            IF Paragraph(FileIndex)
              THEN DoParagraph
              ELSE DoCR;
        END {case};  
        FileIndex := FileIndex+1;
        ChUsed    := ChUsed+1;    
      UNTIL (GoodCh > NbrChPerLine) OR P_Graph;
      {Assert: there are GoodCh printable characters in S.}
      IF P_Graph
        THEN S[0] := CHR(GoodCh)   
        ELSE EditString;
    END {processline};
  
  PROCEDURE ProcessFile;
    {Create a stack consisting of line numbers versus offset in the 
    input file.  Future references can then be made by specifying a 
    line number.  The new lines are of reduced width.}
    VAR 
      LineNo : long_integer;
      Str    : string;
      MyLine : integer;
      
    PROCEDURE Push(x:long_integer);
      {Push x onto a stack.}
      BEGIN
        Stack[StackIndex] := x;
        StackIndex := StackIndex+1;
      END {push};
        
    BEGIN {processfile}
      InIndex := 0;
      StackIndex := 0;
      WHILE InIndex < NumberFileBytes DO
        BEGIN
          Push(InIndex);
          ProcessLine(InIndex,Used,Str);
          InIndex := InIndex+Used;
        END {while};    
      LastLine := StackIndex-1;
    END {processfile};

  PROCEDURE ShowFile;
    {Display the file on the monitor by specifying line numbers.
    The line numbers correspond to lines of reduced width.}
    VAR 
      LineNbr:integer; 
      ch  :char;
      fix : long_integer;  
      ShowLine   : boolean;
      SearchFind : boolean;
      SearchKey  :str255;
      FirstNumber  : boolean;   {Only show dialog box on first user invcation}


    PROCEDURE WideShow;
      {Show a page as an ordinary text editor would, 
      obeying line feeds, tabs,  and carriage returns.}
      VAR LineCt : integer;
      BEGIN 
        LineCt := 0;
        WHILE (LineCt < 24) AND (InIndex < NumberFileBytes) DO
          BEGIN
            Write(InFile^[InIndex]);
            InIndex := InIndex+1;
            IF InFile^[InIndex] = CHR(CR)
              THEN LineCt := LineCt+1;
          END;  
      END {wideshow};
      
    PROCEDURE ShowSearchTarget(Blanks,ThisLine:string);
      {Highlight the phrase found as the result of a search}
      VAR six,i,FirstIx:integer;
      BEGIN
        six := 1;
        {This won't find and highlight if the key is partially
        contained on the next line.  This is, after all, _free_ software.}
        FirstIx := POS(SearchKey,ThisLine);
        IF FirstIx > 0 
          THEN 
            BEGIN
              Write(Blanks);
              FOR six := 1 TO FirstIx-1 DO
                Write(ThisLine[six]);
              InverseVideo;
              FOR six := FirstIx TO FirstIx+LENGTH(SearchKey) DO
                Write(ThisLine[six]);
              NormVideo;
              six := FirstIx + LENGTH(SearchKey);
              FOR i := six TO LENGTH(ThisLine) DO Write(ThisLine[i]);
              WriteLn;
            END
          ELSE WriteLn(Blanks,ThisLine);    
        SearchFind := FALSE;  
      END {showsearchtarget};
      
    PROCEDURE NarrowShow;
      {Show revised text lines.}
      VAR Line:integer; MyStr:string; Spaces:string;
      BEGIN
        {Spaces will be used to center the text on the screen.}
        Spaces := '                                                     ';
        Spaces[0] := CHR(  (80-NbrChPerLine) DIV 2  );
        FOR Line := 1 TO 24 DO
          BEGIN
            ProcessLine(InIndex,Used,MyStr);
            InIndex := InIndex+Used;
            IF (Line = 3) AND SearchFind
              THEN ShowSearchTarget(Spaces,MyStr)
              ELSE WriteLn(Spaces,MyStr);
          END;
      END {narrowshow};

    FUNCTION LowerCase(ch:char):char;
      BEGIN
        IF (ch >= 'A') AND (ch <= 'Z') 
          THEN LowerCase := CHR(ORD(ch) + 32);
      END {lowercase};

    PROCEDURE ProcessNumber(Digit:char);    
      {Let the user type in a number.  If it is successful,
      set LineNbr to this value.  The parameter is the first digit.}
      VAR ch:char; Number:integer;
      
      PROCEDURE ExplainNumber;
        BEGIN
          ClrScrn;
          Find_Dialog(NbrBox,MyBox);
          Center_Dialog(MyBox);
          Show_Dialog(MyBox,NbrBox);
          GoToXY(18,26); Write('This file has ',LastLine,' lines.');
        END {explainnumber};

      FUNCTION GetCh:char;  
        {The user may forget to press return at the end of a number.
        Give him (probably a her) a reasonable time to respond and then
        prompt him/her.}
        VAR Ctr:long_integer; KeyWasPressed:boolean;

        FUNCTION CONSTAT:boolean;
          GEMDOS($0B);

        BEGIN
          Ctr := 0;
          REPEAT
            Ctr := Ctr+1;
          UNTIL ConStat OR (Ctr > 10000);  {About a 4 second delay}
          IF ConStat
            THEN GetCh := ReadNoEcho
            ELSE
              IF NOT FirstNumber
                THEN
                  BEGIN
                    ClrScrn;
                    GoToXY(12,23); Write('Press Return at the end of a number.');
                    GetCh := ReadNoEcho;
                  END;  
        END {getch};
        
      BEGIN {processnumber};
        IF FirstNumber
          THEN ExplainNumber;
        Number := ORD(Digit)-48;
        ch := GetCh;
        WHILE (ch <> CHR(CR)) AND (Number < 10000) DO
          BEGIN
            IF (ch >= '0') AND (ch <= '9')
              THEN Number := Number*10 + ORD(ch)-48;
            ch := GetCh;
          END; 
        {Assert: Number contains the final selection made by the user.}
        LineNbr := Number-1; {user numbers are 1,2,3, ....    not 0,1,2, ...}
        IF LineNbr > LastLine-23
          THEN LineNbr := LastLine-23;
        {Note that LastLine-23 could be less than 0, thus no ELSE.}  
        IF LineNbr < 0 
          THEN LineNbr := 0;      
        FirstNumber := FALSE;  
      END {processnumber};
      
    PROCEDURE DoSearch(SrhKind:TKindSrh; StartLine:integer);
      {Search for an identifier to be supplied by the user.
      Start the search at StartLine.}
      VAR 
        LgthPlus1 : integer;
        fix2      : long_integer;                  
        ch    : char;
        i,six : integer;
        Find  : boolean;
        junk  : integer;

      PROCEDURE GetSearchPattern;
        VAR junk:integer;
        BEGIN
          Find_Dialog(SrchBox,MyBox);
          Center_Dialog(MyBox);
          junk := Do_Dialog(MyBox,0);
          Obj_SetState(MyBox,junk,Normal,TRUE);
          Get_DEdit(MyBox,SrhStrng,SearchKey);
        END {getsearchpattern};
        
      PROCEDURE SearchInnards;
        LABEL 100;
              
      PROCEDURE ShowFind(Adr:long_integer);
        {Convert the offset to a line number and repaint the screen with
        the target line on screen line 3 if possible.}
        VAR lix:integer;
        BEGIN
          lix := 0;  {line index}
          REPEAT
            lix := lix+1;
          UNTIL Stack[lix] > Adr;
          lix := lix-3;
          {Major menu processing loop will repaint the screen.
          We provide the line number of the top line.}
          IF lix >= 0
            THEN LineNbr := lix
            ELSE LineNbr := 0;
          SearchFind := TRUE;
        END {showfind};
        
      PROCEDURE NoFind;
        VAR junk:integer;
        BEGIN  
          junk := Do_Alert('[0][No find][ OK ]',1);
          SearchFind := FALSE;
          GoTo 100;
        END {nofind};
        
        BEGIN {searchinnards}  
          {This block of code is designed for speed.  To get this
          speed it has some tawdry aspects and violates
          some of my notions of good programming practice.}  
          LgthPlus1 := LENGTH(SearchKey)+1;
          ch   := SearchKey[1];
          Find := FALSE;
          REPEAT 
            WHILE (ch <> InFile^[fix]) AND (fix < NumberFileBytes) DO
              fix := fix+1;
            IF fix = NumberFileBytes 
              THEN NoFind;
            {Assert: first letter matches.}  
            fix2 := fix+1;
            six := 2;
            WHILE (SearchKey[six] = InFile^[fix2]) AND (six <= LgthPlus1) DO
              BEGIN
                six  := six+1;
                fix2 := fix2+1;
              END; 
            IF six > LENGTH(SearchKey)
              THEN Find := TRUE
              ELSE fix := fix+1;
          UNTIL Find OR(fix > NumberFileBytes);
          {Assert: If Find is true, fix is the offset in the 
          input file of the desired string.}
          IF Find
            THEN ShowFind(fix) 
            ELSE NoFind;
100:          
        END {searchinnards};

      BEGIN {dosearch}
        SearchFind := FALSE;
        {Convert starting line number into an offset in the input buffer.}
        fix := Stack[StartLine];
        IF SrhKind = FirstSearch 
          THEN 
            BEGIN
              GetSearchPattern;
              IF LENGTH(SearchKey) > 0
                THEN SearchInnards
                ELSE {null search, return to caller}
            END
          ELSE
            SearchInnards;
      END {dosearch};

    PROCEDURE ShowLineNumber;
      BEGIN
        GoToXY(25,28);
        InverseVideo;
        Write(' Top line is line number ',LineNbr+1,' ');
        NormVideo;
        ShowLine := FALSE;
      END {showlinenumber};
                                        
    BEGIN {showfile}
      ShowLine    := FALSE;
      FirstNumber := TRUE;
      LineNbr := 0;
      REPEAT
        InIndex := Stack[LineNbr];
        ClrScrn;
        IF Mode = Wide
          THEN WideShow
          ELSE NarrowShow;
        IF ShowLine
          THEN ShowLineNumber;
        {Wait here for user to tell us what to do next.}
        ch := ReadNoEcho;  
        IF ch = CHR(UpArrow)
          THEN BEGIN IF (LineNbr > 0) THEN LineNbr := LineNbr-1; END
        ELSE IF ch = CHR(DownArrow)
          THEN LineNbr := LineNbr+1
        ELSE IF ch = CHR(Help)
          THEN ShowHelp
        ELSE IF (ch >= '0') AND (ch <= '9')
          THEN ProcessNumber(ch)
        ELSE CASE LowerCase(ch) OF
          'q' : HALT;
          'a' : ShowAbout;
          'b' : IF LineNbr > 22
                  THEN LineNbr := LineNbr-23
                  ELSE LineNbr := 0;
          'f' : LineNbr := 0;
          'l' : LineNbr := LastLine-23;
          'w' : Mode := Wide;
          'n' : Mode := Narrow;
          's' : DoSearch(FirstSearch,LineNbr);
          'r' : DoSearch(RepeatedSearch,LineNbr+4);
          '?' : ShowLine := TRUE;
          'p' : BEGIN
                  REWRITE(OUTPUT,'PRN:');
                  InIndex := Stack[LineNbr];
                  IF Mode = Narrow
                    THEN NarrowShow
                    ELSE WideShow;
                  REWRITE(OUTPUT,'CON:');
                END;  
          OTHERWISE :  IF ((LineNbr+23) < LastLine)
                         THEN LineNbr := LineNbr+23
                         ELSE LineNbr := LastLine-23; 
        END {case};   
      UNTIL ch = 'q';  
    END {showfile};
    
  BEGIN {doit}
    GetFile(InFile);
    AppendEOF_Text;
    Patience;
    Mode := Narrow;
    NbrChPerLine := GetLineWidth;
    ProcessFile;
    ShowHelp;
    ClrScrn;
    {Assert: the entire file has been read and analyzed.
    Each entry in Stack contains the offset in the input file
    of where a 'short' line should start.  For example if Stack[3]
    contained the value 170 it would indicate that the first
    character for line 4 was at offset 170 in the file.}
    ShowFile;
  END {doit};

BEGIN {main}
  REWRITE(Printer,'PRN:');
  ClrScrn;
  IF Init_GEM >= 0
    THEN 
      BEGIN
        IF Load_Resource('NARRWTXT.RSC')
          THEN DoIt
          ELSE 
            BEGIN
              WriteLn('Can''t find .RSC file');
              ReadLn;
              HALT;
            END;  
         Exit_GEM;
      END;  
END {program}.
(*                        TABLE OF CONTENTS
                ( line numbers may vary slightly )
    8 PROGRAM NarrowText;
    9 {$I GEMSUBS.PAS}
   10 {$I NARRWTXT.I}  {Contains CONST declaration}
   27 {$I AUXSUBS.PAS}
   29 FUNCTION ReadNoEcho:char;
   32  |FUNCTION RdNoEcho:long_integer;
   35  |BEGIN {ReadNoEcho}
   43 PROCEDURE DoIt;
   58  |FUNCTION GetLineWidth:integer;
   67  |PROCEDURE GetFile(VAR FileVar:T116);
   71  | |{$I GETPATH.PAS}
   72  | |{$I GETFLSZ.PAS}
   92  |PROCEDURE AppendEOF_Text;
  104  |PROCEDURE Patience;
  111  |PROCEDURE ShowHelp;
  113  | |BEGIN {showhelp}
  122  |PROCEDURE ShowAbout;  
  132  |PROCEDURE ProcessLine(FileIndex:long_integer; 
  145  | |FUNCTION FindLastSpace(LastIx:integer; VAR SpacePtr:integer):boolean;
  157  | |FUNCTION Paragraph(i:long_integer):boolean;  
  182  | |PROCEDURE EditString;
  196  | |FUNCTION CharKind(ch:char):T19;
  206  | |FUNCTION HaveLine:boolean;      
  215  | |PROCEDURE DoParagraph;
  229  | |PROCEDURE DoCR;  
  241  | |PROCEDURE ProcessTextRapidly;  
  260  | |BEGIN {processline}
  307  |PROCEDURE ProcessFile;
  316  | |PROCEDURE Push(x:long_integer);
  323  | |BEGIN {processfile}
  335  |PROCEDURE ShowFile;
  348  | |PROCEDURE WideShow;
  363  | |PROCEDURE ShowSearchTarget(Blanks,ThisLine:string);
  389  | |PROCEDURE NarrowShow;
  406  | |FUNCTION LowerCase(ch:char):char;
  412  | |PROCEDURE ProcessNumber(Digit:char);    
  417  | | |PROCEDURE ExplainNumber;
  426  | | |FUNCTION GetCh:char;  
  432  | | | |FUNCTION CONSTAT:boolean;
  452  | | |BEGIN {processnumber};
  473  | |PROCEDURE DoSearch(SrhKind:TKindSrh; StartLine:integer);
  484  | | |PROCEDURE GetSearchPattern;
  494  | | |PROCEDURE SearchInnards;
  497  | | |PROCEDURE ShowFind(Adr:long_integer);
  515  | | |PROCEDURE NoFind;
  523  | | | |BEGIN {searchinnards}  
  555  | | |BEGIN {dosearch}
  571  | |PROCEDURE ShowLineNumber;
  580  | |BEGIN {showfile}
  630  |BEGIN {doit}
  647 BEGIN {main}
