/*
**********************************************************************
**
** KOUNT  count lines, words and chars with/without newlines in
**        files supplied via command line, and produce an
** individual file and total of all files report. The program
** is a modification of the WC program by Kerninghan and Ritchie.
** It is in the Public Domain.
**
** BATCH.TTP directives (link):
**
** link68 [u] kount.68k=gemstart,kount,osbind.o,libf,libm,gemlib
** relmod kount
** rm kount.68k
** wait
**
** Usage: KOUNT [-S] fil1 [fil2[filn...]]
**          
** 10/04/86 A new version that correctly uses longs thanks to
**          some help from B.I. in the great white north 
**          "Thanks eh!" Also this version does not use any of my
**          silly #defines that make C look like FORTRAN (well
**          sort of like FORTRAN).
** 10/05/86 Another new version that checks for a -S on the command
**          line. If there, the program won't wait for a keystroke
**          before exit as th option means we are running under a 
**          shell of some sort.
**
**********************************************************************
*/

#include "tosdefs.h"
#include "osbind.h"
#include "stdio.h"

main(argc, argv) 

/*
** Parameter definitions (command line)
*/

char *argv[];
int argc;

{
   /*
   ** Variable declarations
   */

   FILE *fp,       /*      Stream                        */
        *fopen();  /*      File open routine             */
 
   char *next;     /*      next argument character       */

   long int l,     /* l  = lines found                   */
       c,          /* c  = characters found              */
       n,          /* n  = characters found w/nl         */
       w,          /* w  = words found                   */
       i,          /* i  = file(s) counted               */
       lt,         /* lt = total lines found             */
       ct,         /* ct = total characters found        */
       nt,         /* nt = total characters found w/nl   */
       wt,         /* wt = total words found             */
       t,          /* t  = files from command line count */
       igot;       /*      console input                 */

   int f,          /* f  = transition flag               */
       s,          /* s  = shell flag                    */
       d;          /* d  = current character             */ 

   /*
   ** Initialize the variables
   */

   f = s = NO;
   l = c = n = w = lt = ct = nt = wt = t = i = 0;

   /*
   ** Check for arguments, if none explain how we work and quit
   */

   if (argc <= 1) {
      printf ("\007\007"); /* Ding! */            
      printf ("usage: KOUNT [-S] file1 [file2 [...filen]]\n");
      printf ("       -S indicates a Shell is being used\n");
      printf ("Press any key to continue.");
      igot = Cconin ();
      printf ("\n");
      exit (0);
   }
   
   /* 
   ** Take care of introductions and let em' know this is PD
   */

   printf ("KOUNT 1.03 10/05/86 by Dan Rhea. Public Domain.\n");

   /* 
   ** Now start grabbing filenames from the command line 
   */

   while (--argc > 0) { 

      /*
      ** Bump the file counter and open the file
      */

      ++i;
      if ((fp = fopen(*++argv, "r")) <= NULL) {

         /*
         ** If we couldn't open it, tell the user, back up the 
         ** file count and jump to the next file
         ** Also is a -S is there than we are in a shell so just
         ** set a flag but show no error
         */

         if ((*argv)[0] == '-') {
            next = argv[0]+1;
            if ((*next == 'S') || (*next == 's'))
               s = YES;
            else {
               printf ("\007\007Error:"); /* Ding! */
               printf (" Invalid option: %c.\n",*next);
            }
         }
         else {
            printf ("\007\007Error:"); /* Ding! */
            printf (" Unable to open %s. fp = %ld.\n",*argv,fp);
         }
         --i;
         continue;
      }

      /* 
      ** Scan the file till we hit EOF
      */

      while ((d = getc(fp)) != EOF) { 

         /*
         ** Count the characters 
         */

         if (d != '\n') {  
            ++c, ++n, ++ct, ++nt;
         }
         
         /* 
         ** Count the lines
         */

         else if (d == '\n') {
            ++l, ++n, ++lt, ++nt;
         }
         
         /*
         ** Check for the end of a word and count it if we think
         ** we found one (end of a word is a newline, a tab or 
         ** any white space).
         */

         if (d == ' ' || d == '\n' || d == '\t')
            f = NO;
         else if (f == NO) { 
            f = YES, ++w, ++wt;
         }
      }

      /*
      ** Show the totals for this file (if we found anything)
      ** then close the file and prepare for the next victim
      */

      if (l != 0) {
         printf ("%-16s: ",*argv);
         printf ("%7ld Lines, %7ld Words, ",l,w);
         printf ("%7ld Chars, %7ld W/nl.\n",c,n);
         l = c = n = w = 0;
         ++t;
         fclose (fp);
      }
   }

   /*
   ** if more than one file was counted than show the grand
   ** totals. Then hang around till the user hits a key before
   ** returning to desktop or shell. Unless -S was selected
   */

   if (t > 1) {
      printf ("%6ld Files: ",i);
      printf ("%8ld Lines, %8ld Words, ",lt,wt);
      printf ("%8ld Chars, %8ld W/nl.\n",ct,nt);
   }
   if (s == NO) {
      printf ("Press any key to continue.");
      igot = Cconin ();
      printf ("\n"); 
   }

} /* The end of the KOUNT program! */ 
